/**
  *  \file client/dialogs/shipbuilddlg.cc
  *  \brief WShipBuildDialog dialog base
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/shipbuilddlg.h"
#include "client/dlg-component.h"
#include "client/widgets/specview.h"
#include "game/global.h"
#include "game/shipquery.h"
#include "gfx/palette.h"
#include "resmgr/resid.h"
#include "resmgr/resmgr.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/framegroup.h"
#include "util/nls.h"
#include "client/dialogs/buildbilldlg.h"

/** Constructor.
    \param name Heading for dialog
    \param planet Planet to work on
    \param withPart true to offer build part controls
    \param hulls Hulls to offer for building (will be cleared) */
WShipBuildDialog::WShipBuildDialog(string_t name, GPlanet& planet, bool withPart, std::vector<int>& hulls)
    : UIWindow(name, UIWindow::BLUE_WINDOW, COLOR_DARK),
      hull_list(3000, 10, 15 * font_ems[FONT_NORMAL], &planet),
      engine_list(3001, 12 * font_ems[FONT_NORMAL], &planet),
      beam_list(3002, 12 * font_ems[FONT_NORMAL], &planet),
      torp_list(3003, 12 * font_ems[FONT_NORMAL], &planet),
      hull_storage(withPart),
      torp_storage(withPart),
      beam_storage(withPart),
      beam_count(false),
      torp_count(true),
      withPart(withPart),
      action(planet),
      item_cost_dpy(planet, false),
      cards(),
      tabs(cards),
      conn_update(planet.sig_changed.add(this, &WShipBuildDialog::onPlanetChanged))
{
    init(hulls);
}

void
WShipBuildDialog::init(std::vector<int>& hulls)
{
    setLayout(UIVBoxLayout::instance5);

    add(h.add(new UIFocusIterator(1000, 1003, UIFocusIterator::fi_Horiz)));

    tabs.addTab(1000, UIKeyString(_("Starship Hulls")));
    tabs.addTab(1001, UIKeyString(_("Engines")));
    tabs.addTab(1002, UIKeyString(_("Beams")));
    tabs.addTab(1003, UIKeyString(_("Torpedoes")));

    /* Hulls */
    UIGroup& hull_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1000));
    hull_group.add(UIFrameGroup::wrap(h, hull_list));
    hull_group.add(hull_storage);
    hull_group.add(hull_info);
    hull_list.swapContents(hulls);
    hull_list.scrollToHull(action.getBuildOrder().hull);
    hull_list.sig_moved.add(this, &WShipBuildDialog::updateHullPage);

    /* Engines */
    UIGroup& engine_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1001));
    UIGroup& engine_vgroup = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& engine_hgroup = h.add(new UIGroup(UIHBoxLayout::instance5));
    engine_group.add(UIFrameGroup::wrap(h, engine_list));
    engine_group.add(engine_diag);
    engine_vgroup.add(engine_info);
    engine_vgroup.add(h.add(new UISpacer()));
    engine_vgroup.add(engine_storage);
    engine_vgroup.add(engine_hgroup);
    if (withPart) {
        engine_hgroup.add(h.add(new UIButton(_("Spc-Build"), ' ')));
    }
    engine_hgroup.add(h.add(new UISpacer()));
    engine_group.add(engine_vgroup);
    engine_list.scrollTo(action.getBuildOrder().engine - 1, 1);
    engine_list.sig_moved.add(this, &WShipBuildDialog::updateEnginePage);

    /* Beams */
    UIGroup& beam_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1002));
    UIGroup& beam_vgroup = h.add(new UIGroup(UIVBoxLayout::instance0));
    beam_group.add(UIFrameGroup::wrap(h, beam_list));
    beam_group.add(beam_storage);
    beam_vgroup.add(beam_info);
    beam_vgroup.add(h.add(new UISpacer()));
    beam_vgroup.add(beam_count);
    beam_group.add(beam_vgroup);
    beam_list.scrollTo(action.getBuildOrder().beam_type - 1, 1);
    beam_list.sig_moved.add(this, &WShipBuildDialog::updateBeamPage);

    /* Torps */
    UIGroup& torp_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1003));
    UIGroup& torp_vgroup = h.add(new UIGroup(UIVBoxLayout::instance0));
    torp_group.add(UIFrameGroup::wrap(h, torp_list));
    torp_group.add(torp_storage);
    torp_vgroup.add(torp_info);
    torp_vgroup.add(h.add(new UISpacer()));
    torp_vgroup.add(torp_count);
    torp_group.add(torp_vgroup);
    torp_list.scrollTo(action.getBuildOrder().torp_type - 1, 1);
    torp_list.sig_moved.add(this, &WShipBuildDialog::updateLauncherPage);

    cards.add(hull_group);
    cards.add(engine_group);
    cards.add(beam_group);
    cards.add(torp_group);
    cards.sig_focus_changed.add(this, &WShipBuildDialog::updateBuildOrder);

    add(tabs);
    add(cards);

    /* Bill */
    UIGroup& bill_group = h.add(new UIGroup(UIHBoxLayout::instance5));
    bill_group.add(item_cost_dpy);
    bill_group.add(order_dpy);
    add(bill_group);

    /* Initialize GUI */
    hull_list.setFocusedWidget();
    updateForNewHull(true);
    updateEnginePage();
    updateBeamPage();
    updateLauncherPage();

    beam_count.sig_count_changed.add(this, &WShipBuildDialog::updateBuildOrder);
    torp_count.sig_count_changed.add(this, &WShipBuildDialog::updateBuildOrder);
}

WShipBuildDialog::~WShipBuildDialog()
{ }

bool
WShipBuildDialog::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && event.key.code == ' ') {
        GPlanet& p = action.getPlanet();
        switch (tabs.getCurrentTab()) {
         case 1000:
            if (int slot = getTruehullSlot(p.getOwner(), hull_list.getCurrentHull()))
                doBuildComponent(getDisplayedTurn(), p.getId(),
                                 tl_Hull, slot,
                                 getHull(hull_list.getCurrentHull()));
            break;
         case 1001:
            doBuildComponent(getDisplayedTurn(), p.getId(),
                             tl_Engine, engine_list.getCurrent()+1,
                             getEngine(engine_list.getCurrent()+1));
            break;
         case 1002:
            doBuildComponent(getDisplayedTurn(), p.getId(),
                             tl_Beam, beam_list.getCurrent()+1,
                             getBeam(beam_list.getCurrent()+1));
            break;
         case 1003:
            doBuildComponent(getDisplayedTurn(), p.getId(),
                             tl_Torpedo, torp_list.getCurrent()+1,
                             getTorp(torp_list.getCurrent()+1));
            break;
        }
        return true;
    }
    if (event.type == ev_Key && event.key.code == 's' && tabs.getCurrentTab() == 1000) {
        /* 's' on hull tab: build ship query */
        GShipQuery q;
        q.setHullId(hull_list.getCurrentHull());
        q.setPlayerDisplaySet(GPlayerSet(action.getPlanet().getOwner()));
        q.setEngineId(engine_list.getCurrent()+1);
        q.setOwner(action.getPlanet().getOwner());
        q.complete(getDisplayedTurn().getCurrentUniverse());

        /* build dialog */
        WSpecView sv;
        sv.setQuery(q);
        sv.doStandardDialog();
        return true;
    } else {
        return UIWindow::handleEvent(event, second_pass);
    }
}

/** Update for a new hull selection. Updates the hull page and dependant
    values on the other pages.
    \param initial true: This is the initial invokation, keep less-than-max weapons.
                   false: This is a selection change, initialize to max weapons. */
void
WShipBuildDialog::updateForNewHull(bool initial)
{
    const int hull = hull_list.getCurrentHull();
    const GPlanet& p = action.getPlanet();

    hull_info.setHullId(hull);
    if (int n = getTruehullSlot(p.getOwner(), hull))
        hull_storage.setCount(p.getBaseHullStoreSlot(n));
    else
        hull_storage.setCount(0);
    hull_storage.setImage(requestPixmap(ResId(res::ship, getHull(hull).getInternalPictureNumber(), hull)));

    engine_storage.setEnginesNeeded(getHull(hull).getNumEngines());
    beam_count.setPartsMax(getHull(hull).getMaxBeams());
    torp_count.setPartsMax(getHull(hull).getMaxLaunchers());

    if (initial && hull == action.getBuildOrder().hull) {
        beam_count.setPartsWanted(action.getBuildOrder().beam_count);
        torp_count.setPartsWanted(action.getBuildOrder().torp_count);
    }

    updateBuildOrder();
}

/** Update hull page after change to hull selection. */
void
WShipBuildDialog::updateHullPage()
{
    updateForNewHull(false);
}

/** Update beam page after change to beam selection. */
void
WShipBuildDialog::updateBeamPage()
{
    const int beam = beam_list.getCurrent()+1;
    beam_info.setBeamId(beam);
    beam_storage.setCount(action.getPlanet().getBaseBeamStore(beam));
    beam_storage.setImage(requestPixmap(ResId("beam", beam)));
    updateBuildOrder();
}

/** Update launcher page after change to launcher selection. */
void
WShipBuildDialog::updateLauncherPage()
{
    const int torp = torp_list.getCurrent()+1;
    torp_info.setTorpId(torp);
    torp_storage.setCount(action.getPlanet().getBaseLauncherStore(torp));
    torp_storage.setImage(requestPixmap(ResId("launcher", torp)));
    updateBuildOrder();
}

/** Update engine page after change to engine selection. */
void
WShipBuildDialog::updateEnginePage()
{
    const int engine = engine_list.getCurrent()+1;
    const GPlanet& p = action.getPlanet();
    engine_info.setEngineId(engine);
    engine_diag.setEngine(engine);
    engine_storage.setEnginesAvailable(p.getBaseEngineStore(engine));
    updateBuildOrder();
}

/** Update build order from selection. This also updates all cost displays. */
void
WShipBuildDialog::updateBuildOrder()
{
    TBuildOrder bo;

    bo.hull   = hull_list.getCurrentHull();
    bo.engine = engine_list.getCurrent()+1;
    bo.beam_type  = beam_list.getCurrent()+1;
    bo.beam_count = beam_count.getPartsWanted();
    bo.torp_type  = torp_list.getCurrent()+1;
    bo.torp_count = torp_count.getPartsWanted();
    bo.unused = 0;
    action.setBuildOrder(bo);
    order_dpy.setBuildOrder(bo);
    item_cost_dpy.setTotalCost(action.getCost());

    // since this might be called during destruction, we cannot use
    // 'tabs.getCurrentTab()'. 'tabs' might already be gone.
    if (UIBaseWidget* p = cards.getFocusedChild()) {
        switch (p->getId()) {
         case 1000: {
            item_cost_dpy.setPartCost(getHull(hull_list.getCurrentHull()).getCost());
            item_cost_dpy.setPartTech(action.getPlanet().getBaseTechLevel(tl_Hull),
                                      getHull(hull_list.getCurrentHull()).getTech());
            break;
         }
         case 1001: {
            GEngine& e = getEngine(engine_list.getCurrent()+1);
            item_cost_dpy.setPartCost(e.getCost());
            item_cost_dpy.setPartTech(action.getPlanet().getBaseTechLevel(tl_Engine), e.getTech());
            break;
         }
         case 1002: {
            GBeam& b = getBeam(beam_list.getCurrent()+1);
            item_cost_dpy.setPartCost(b.getCost());
            item_cost_dpy.setPartTech(action.getPlanet().getBaseTechLevel(tl_Beam), b.getTech());
            break;
         }
         case 1003: {
            GTorpLauncher& t = getTorp(torp_list.getCurrent()+1);
            item_cost_dpy.setPartCost(t.getCost());
            item_cost_dpy.setPartTech(action.getPlanet().getBaseTechLevel(tl_Torpedo), t.getTech());
            break;
         }
        }
    }
}

void
WShipBuildDialog::onPlanetChanged()
{
    item_cost_dpy.setTotalCost(action.getCost());
    if (UIBaseWidget* p = cards.getFocusedChild()) {
        switch (p->getId()) {
         case 1000:
            updateHullPage();
            hull_list.drawWidget(false);
            break;
         case 1001:
            updateEnginePage();
            engine_list.drawWidget(false);
            break;
         case 1002:
            updateBeamPage();
            beam_list.drawWidget(false);
            break;
         case 1003:
            updateLauncherPage();
            torp_list.drawWidget(false);
            break;
        }
    }
}

void
WShipBuildDialog::onDetailedBill()
{
    doBuildDetailedBill(action.getPlanet(), action);
}

void
WShipBuildDialog::setBuildOrder(const TBuildOrder& order)
{
    // Prepare it
    TBuildOrder o = order;
    action.prepareBuildOrder(o, action.getPlanet());

    // Set into action
    action.setBuildOrder(o);

    // Set into GUI
    hull_list.scrollToHull(o.hull);
    engine_list.scrollTo(o.engine - 1, 1);
    beam_list.scrollTo(o.beam_type - 1, 1);
    torp_list.scrollTo(o.torp_type - 1, 1);
    beam_count.setPartsWanted(o.beam_count);
    torp_count.setPartsWanted(o.torp_count);
}
