/**
  *  \file game/actions/cargocost.h
  *  \brief Cargo Cost Transaction
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_ACTIONS_CARGOCOST_H
#define PCC_V2_GAME_ACTIONS_CARGOCOST_H

#include "cpluslib/types.h"
#include "cpluslib/sigslot.h"
#include "game/cost.h"
#include "game/cargotype.h"
#include "game/transact.h"
#include "game/error.h"

class GCargoContainer;

/** \class GCargoCostTransaction
    \brief General Cost Transaction

    This is the base class for all transactions which cost some
    resources, for example, buying structures, building a base or
    raising tech. It provides functions to change the bill, and do the
    allocation on a cargo container.

    This class supports costs consisting of the three industry
    minerals (T/D/M), money and supplies (GCost).

    Cargo is billed to a GCargoContainer which usually is a planet.
    It must be part of the same GMultiTransaction as this
    GCargoCostTransaction, and committed/canceled with it.
    A derived class will override the commit() method to actually
    perform the desired action with the commit. */
class GCargoCostTransaction : public GTransaction {
    GCargoContainer& container;

    GCost cost;
    GCost reserved;

    void update();

 protected:
    void changeCost(GCargoType type, int32_t amount);
    void changeCost(const GCost& cost);
    void setCost(GCargoType type, int32_t amount);
    void setCost(const GCost& cost);

    void commit();
    void cancel();

 public:
    GCargoCostTransaction(GCargoContainer& container);
    ~GCargoCostTransaction();

    bool isValidCost() const;

    GCost   getCost() const;
    int32_t getCost(GCargoType type) const;
    int32_t getRemainingCargo(GCargoType type) const;
    int32_t getMissingCargo(GCargoType type) const;
    int32_t getReservedCargo(GCargoType type) const;

    GError getError();

    GCargoContainer& getContainer() const;

    Signal0 sig_cost_changed;
};

/** Access to container used by this transaction. */
inline GCargoContainer&
GCargoCostTransaction::getContainer() const
{
    return container;
}

#endif
