/**
  *  \file game/load-rst.cc
  *  \brief RST Loading
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/load.h"
#include "game/rst.h"
#include "game/global.h"
#include "util/console.h"
#include "util/nls.h"
#include "util/except.h"
#include "cpluslib/format.h"
#include "game/stat.h"
#include "game/racename.h"
#include "io/streams.h"
#include "game/classicvcr.h"
#include "game/backup.h"
#include "game/gamedirs.h"
#include "util/conf.h"

/** Load a RST file. */
void
loadResult(GGameTurn& trn, Stream& s, int player, bool withBackup)
{
    TInt16 n;
    TResultGen gen;

    console.write(LOG_INFO, format(_("Loading %s RST file..."),
                                   player_racenames.getAdjName(player)));

    GResultFile rst(s);
    GPlayerSet  src(player);

    s.seek(rst.getSectionOffset(GResultFile::sec_Gen));
    getStructureT(s, gen);
    if (gen.player_id != player)
        throw FileFormatException(s, format(_("File is owned by player %d, should be %d"),
                                            gen.player_id, player));
    trn.setHaveData(player);
    trn.setGen(player, GGen(gen));
    getStatistics().mergeGen(trn.getGen(player));

    /* Backup */
    if (withBackup) {
        s.seek(0);
        GBackupFileTemplate tpl;
        ProgressMonitor pm;
        tpl.setPlayerNumber(player);
        tpl.setTurnNumber(gen.turn);
        tpl.setGameDirectory(game_dir_name);
        tpl.copyFile(opt_BackupResult(), s, pm);
    }

    /* Ships */
    s.seek(rst.getSectionOffset(GResultFile::sec_Ships));
    getStructureT(s, n);
    loadShips(trn, s, n.value, LoadBoth, src);

    /* Targets */
    s.seek(rst.getSectionOffset(GResultFile::sec_Targets));
    getStructureT(s, n);
    loadTargets(trn, s, n.value, TargetPlaintext, LoadCurrent, src);

    /* Planets */
    s.seek(rst.getSectionOffset(GResultFile::sec_Planets));
    getStructureT(s, n);
    loadPlanets(trn, s, n.value, LoadBoth, src);

    /* Starbases */
    s.seek(rst.getSectionOffset(GResultFile::sec_Bases));
    getStructureT(s, n);
    loadBases(trn, s, n.value, LoadBoth, src);

    /* Messages */
    s.seek(rst.getSectionOffset(GResultFile::sec_Messages));
    trn.getInbox().loadInbox(s, player);

    /* SHIPXY */
    s.seek(rst.getSectionOffset(GResultFile::sec_ShipXY));
    loadShipXY(trn, s, (rst.getSectionOffset(GResultFile::Section(GResultFile::sec_ShipXY+1))
                        - rst.getSectionOffset(GResultFile::sec_ShipXY)),
               LoadBoth,
               src,
               GPlayerSet());

    /* VCRs */
    s.seek(rst.getSectionOffset(GResultFile::sec_Vcrs));
    trn.setVcrs(new GClassicVcrDatabase(s));

    /* Windows part of RST */
    if (rst.haveSection(GResultFile::sec_Kore)) {
        /* KORE */
        s.seek(rst.getSectionOffset(GResultFile::sec_Kore));
        loadKoreMinefields(trn, s, 500, player);
        loadKoreIonStorms(trn, s, 50);
        loadKoreExplosions(trn, s, 50);
        player_racenames.load(s); /* FIXME: configurable? */
        host_racenames = player_racenames;
        loadUfos(trn, s, 1, 100);

        char buffer[8];
        if (s.read(buffer, 8) == 8 && std::memcmp(buffer, "1120", 4) == 0) {
            int32 count = getInt32(buffer, 4);
            if (count < 0 || count > NUM_SHIPS)
                throw FileFormatException(s, _("Unbelievable number of visual contacts"));
            loadTargets(trn, s, count, TargetEncrypted, LoadCurrent, src);
        }
    }

    if (rst.haveSection(GResultFile::sec_Skore)) {
        /* SKORE */
        TInt16 count;
        s.seek(rst.getSectionOffset(GResultFile::sec_Skore));
        if (getStructure(s, count) && count.value > 100) {
            loadUfos(trn, s, 101, count.value - 100);
        }
    }
}
