/**
  *  \file game/transact.cc
  *  \brief Transactions
  *
  *  This provides the basic transaction class, on which most cargo
  *  manipulation stuff is based.
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/transact.h"
#include "cpluslib/assert.h"

/*! \class GTransaction

    This class supports user-controlled transactions. The transaction
    model is as follows:
    - we start the transaction by constructing a GTransaction descendant
    - the user prepares the transaction by calling some of the
      descentants methods
    - when the user hits "OK", the transaction is committed, otherwise
      it is cancelled.

    If the transaction is asked to do something impossible, it throws
    a GError. By convention, this happens at two places:
    - constructor
    - commit() in error state

    Users call doCancel()/doCommit() to cancel or commit the
    transaction. doCommit() will automatically query the error state
    and make sure commit() is only called in a successful TA.
    GTransaction descendants must override the cancel/commit methods
    to do the actual work. Users must not call cancel()/commit()!

    Note that a transaction must always be cleanly committed or
    cancelled, it is usually an error to destroy a transaction via a
    destructor without having called one of doCommit() or doCancel().
    There should be exactly one commit() or cancel(), it is an error
    to call both for the same transaction. To ensure that, include the
    sequence
    <pre>if(!isFinished()) doCancel();</pre>
    in your destructor.

    Our transactions can be nested, overlapped and all other nasty
    things. We do not support pre-emptive concurrency, though. (Read:
    GTransaction does not contain any means of inter-object
    communication which could help or hurt you.)

    In a multi-part transaction, things may depend on each other. For
    example, a sub-transaction may state "if this transaction is
    committed, another sub-transaction must be committed as well". To
    ensure such things, you can use mustCommit() and mustCancel();
    these will trigger an assert fault when the transaction is not
    committed.

    Although it is possible to implement multi-part transactions by
    simply constructing a set of GTransaction descendants, it is
    simpler and safer to use GMultiTransaction as arbiter. */

/** Destructor. */
GTransaction::~GTransaction()
{ }

void
GTransaction::finish(Result r)
{
    ASSERT(must_result == Nothing || must_result == r);
    result = r;
}

/** Make sure this transaction is cancelled. Triggers an assert
    failure when the transaction is committed instead. */
void
GTransaction::mustCancel()
{
    ASSERT(must_result != Committed && result != Committed);
    must_result = Cancelled;
}

/** Make sure this transaction is committed. Triggers an assert
    failure when the transaction is cancelled instead. */
void
GTransaction::mustCommit()
{
    ASSERT(must_result != Cancelled && result != Cancelled);
    must_result = Committed;
}

/** Get error message.
    \default returns no error */
GError
GTransaction::getError()
{
    return GError();
}

/** Commit this transaction, external entry. Makes sure the
    transaction is valid, and then calls commit().
    \throw GError as returned by getError(), if that is an error. */
void
GTransaction::doCommit()
{
    if (GError e = getError()) {
        doCancel();
        throw e;
    }
    commit();
    finish(Committed);
}

/** Cancel this transaction, external entry. */
void
GTransaction::doCancel()
{
    cancel();
    finish(Cancelled);
}

/*************************** GMultiTransaction ***************************/

/*! \class GMultiTransaction

    This class provides a means to support transactions consisting of
    multiple subtransactions in an exception-safe and simple manner.

    It is used as
    \verbatim
      GMultiTransaction m;
      MyOne& t1 = m.addNewSubtransaction(new MyOne());
      MyTwo& t2 = m.addNewSubtransaction(new MyTwo());
    \endverbatim
    because this automatically gets construction and destruction
    right. If an exception happens somewhere inbetween, the
    transaction is automatically cancelled.

    All sub-transactions must be committed or cancelled en bloc using
    our commit() or cancel() functions. */

/** Constructor. */
GMultiTransaction::GMultiTransaction()
    : children()
{ }

/** Destructor. Cancels all transaction in case this multi-part
    transaction is not yet finished. */
GMultiTransaction::~GMultiTransaction()
{
    if (!isFinished()) {
        doCancel();
    }
}

/** Commit. This commits all subtransactions. */
void
GMultiTransaction::commit()
{
    for (std::size_t i = 0; i < children.size(); ++i) {
        children[i]->doCommit();
    }
}

/** Cancel. This cancels all subtransactions. */
void
GMultiTransaction::cancel()
{
    for (std::size_t i = 0; i < children.size(); ++i) {
        children[i]->doCancel();
    }
}

/** Get error message. If any transaction is in error state, returns
    its error message. */
GError
GMultiTransaction::getError()
{
    for (std::size_t i = 0; i < children.size(); ++i) {
        if (GError e = children[i]->getError()) {
            return e;
        }
    }
    return GError();
}
