/**
  *  \file int/export/expjson.cc
  *  \brief Export to JSON file
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "int/export/expjson.h"
#include "int/value.h"
#include "io/textfile.h"
#include "cpluslib/utf8.h"
#include "cpluslib/format.h"
#include "int/compound.h"
#include "int/data.h"
#include "int/args.h"

static bool writeValue(TextFile& tf, IntValue* value, int depth);

static void
writeQuotedString(TextFile& tf, string_t s)
{
    tf.write("\"");
    Utf8Reader rdr(s);
    while (rdr.hasMore()) {
        uint16_t ch = rdr.get();
        if (ch < 32 || ch >= 127) {
            tf.write(format("\\u%04X", ch));
        } else if (ch == '\\' || ch == '\"') {
            tf.write(format("\\%c", ch));
        } else {
            tf.write(format("%c", ch));
        }
    }
    tf.write("\"");
}

static bool
tryWriteArray(TextFile& tf, IntValue* value, int depth)
{
    // Sufficient depth to recurse?
    --depth;
    if (depth <= 0) {
        return false;
    }

    // Correct type?
    IntIndexableValue* iv = dynamic_cast<IntIndexableValue*>(value);
    if (!iv) {
        return false;
    }

    // Has any dimensions?
    if (iv->getDimension(0) == 0) {
        return false;
    }

    // OK, looks like an array. Write as one.
    int32_t dim = iv->getDimension(1);
    tf.write("[");
    for (int32_t i = 1; i <= dim; ++i) {
        if (i != 1) {
            tf.write(",");
        }
        try {
            IntDataSegment argSeg;
            argSeg.pushNew(makeIntValue(i));
            IntArgBlock argBlock(argSeg, 0, 1);

            std::auto_ptr<IntValue> element(iv->get(argBlock));
            writeValue(tf, element.get(), depth);
        }
        catch (...) {
            writeQuotedString(tf, "#<error>");
        }
    }
    tf.write("]");
}

static bool
writeValue(TextFile& tf, IntValue* value, int depth)
{
    if (value == 0) {
        tf.write("null");
    } else if (IntIntValue* iv = dynamic_cast<IntIntValue*>(value)) {
        if (iv->isBool()) {
            tf.write(iv->getValue() ? "true" : "false");
        } else {
            tf.write(iv->toString(true));
        }
    } else if (tryWriteArray(tf, value, depth)) {
        // ok
    } else {
        writeQuotedString(tf, value->toString(false));
    }
}

IntJsonTextExporter::IntJsonTextExporter(TextFile& tf)
    : tf(tf),
      firstField(true),
      firstRecord(true)
{ }

void
IntJsonTextExporter::startTable(const IntExportFieldList& /*fields*/, const TIntTypeHint /*types*/[])
{
    // Start an array
    tf.write("[");
    firstRecord = true;
}

void
IntJsonTextExporter::startRecord()
{
    // Start a hash
    if (!firstRecord) {
        tf.write(",\n");
    }
    tf.write("{");
    firstField = true;
    firstRecord = false;
}

void
IntJsonTextExporter::addField(IntValue* value, const string_t& name, TIntTypeHint /*type*/)
{
    // Write a value
    if (!firstField) {
        tf.write(",\n");
    }
    writeQuotedString(tf, name);
    tf.write(":");
    writeValue(tf, value, 3);
    firstField = false;
}

void
IntJsonTextExporter::endRecord()
{
    tf.write("}");
}

void
IntJsonTextExporter::endTable()
{
    tf.write("]\n");
}
