/**
  *  \file int/fusion.cc
  *  \brief Interpreter: Instruction Fusion
  *
  *  To avoid unnecessary creation of temporary values, we can fuse
  *  instructions together. The idea is to detect patterns, and mark
  *  fusable instructions with a special major opcode. The interpreter
  *  then executes these major opcodes in one cycle, saving
  *  temporaries where possible.
  *
  *  Instruction fusion is an internal optimisation. Fused
  *  instructions are not saved to the VM file; see
  *  IntOpcode::getExternalMajor().
  *
  *  No changes are done to the instruction stream other than the
  *  major opcode. Thus, all information can still be found at the
  *  usual places. In particular, if a jump exists into the middle of
  *  a fused instruction, it will simply proceed by executing the
  *  original unfused instruction.
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/fusion.h"
#include "int/bytecode.h"

/***************************** IntLocalTracer ****************************/

static const uint32_t STORE_DEPTH = 10;

namespace {
    /** Local Variable Tracer.
        Used to find local variables that are guarenteed to be overwritten.
        This is an object to allow re-using some state between different invocations. */
    class IntLocalTracer {
     public:
        IntLocalTracer(const IntBytecodeObject& bco);

        bool isOverwrittenLocal(IntBytecodeObject::pc_t pc, uint16_t address, uint32_t depth);
        bool hasExceptionHandling();

     private:
        const IntBytecodeObject& bco;
        enum { DontKnow, No, Yes } ehState;
    };
}

/** Constructor.
    \param bco BCO to analyze */
IntLocalTracer::IntLocalTracer(const IntBytecodeObject& bco)
    : bco(bco),
      ehState(DontKnow)
{ }

/** Check for overwritten local variable.
    If we know a local variable is overwritten, we can clobber it during execution.
    This function must return true if it can prove that the variable is overwritten,
    false if it is re-used or it cannot be proven.
    To limit the amount of work to do, we trace only as many instructions as given
    by the %depth parameter. If we cannot prove overwriting within so-many instructions,
    we assume the variable is still needed and therefore return false.

    \param pc Start at this program counter
    \param address Look for this local variable
    \param depth Check this many instructions

    \retval true The variable is guaranteed to be overwritten and losing its value is ok
    \retval false The variable's value might still be required, or we are unsure */
bool
IntLocalTracer::isOverwrittenLocal(IntBytecodeObject::pc_t pc,
                                   uint16_t address,
                                   uint32_t depth)
{
    /* Check depth */
    while (depth > 0 && pc < bco.getNumInstructions()) {
        const IntOpcode& op = bco(pc++);
        --depth;       
        switch (op.major) {
         case IntOpcode::maPush:
         case IntOpcode::maFusedUnary:
         case IntOpcode::maFusedBinary:
         case IntOpcode::maFusedComparison2:
         case IntOpcode::maInplaceUnary:
            /* Accept only pushloc for different locals, or literals */
            if ((op.minor == IntOpcode::sLocal && op.arg != address)
                || op.minor == IntOpcode::sLiteral
                || op.minor == IntOpcode::sInteger
                || op.minor == IntOpcode::sBoolean)
            {
                /* accept */
            } else {
                /* reject */
                return false;
            }
            break;

         case IntOpcode::maBinary:
         case IntOpcode::maFusedComparison:
         case IntOpcode::maUnary:
         case IntOpcode::maTernary:
            /* These can throw. Do not accept if it smells like exception handling,
               because the exception handler might want to examine this local variable. */
            if (hasExceptionHandling()) {
                return false;
            }
            break;

         case IntOpcode::maJump:
            /* Trace regular jumps */
            if (op.isLabel()) {
                /* labels: skip */
            } else if (op.isRegularJump()) {
                /* regular jump: continue tracing at jump target */
                if ((op.minor & IntOpcode::jAlways) != IntOpcode::jAlways) {
                    /* conditional jump: trace following instruction first */
                    if (!isOverwrittenLocal(pc, address, depth)) {
                        return false;
                    }
                }
                pc = bco.getJumpTarget(op.minor, op.arg);
            } else {
                /* jdz, catch: too complex/rare to reason about */
                return false;
            }
            break;
            
         case IntOpcode::maIndirect:
         case IntOpcode::maMemref:
            /* These can do anything. Do not accept. */
            return false;

         case IntOpcode::maStack:
            /* Always safe, does not modify locals */
            break;

         case IntOpcode::maPop:
         case IntOpcode::maStore:
            /* These are what we want. If we found the store to our address, report so.
               Otherwise, continue tracing. (We're looking for guaranteed matches;
               unlike for push, finding a possible match is not sufficient.) */
            if (op.minor == IntOpcode::sLocal && op.arg == address) {
                return true;
            }
            break;

         case IntOpcode::maDim:
            /* Always safe, does not modify existing locals */
            break;

         case IntOpcode::maSpecial:
            /* These can do anything */
            return false;
        }
    }

    /* When we're here, we cannot prove that this local is overwritten */
    return false;
}

/** Check whether exception handling is in use.
    \retval true This BCO uses exception handling
    \retval false This BCO does not use exception handling */
bool
IntLocalTracer::hasExceptionHandling()
{
    if (ehState == DontKnow) {
        ehState = No;
        for (IntBytecodeObject::pc_t i = 0, n = bco.getNumInstructions(); i < n; ++i) {
            const IntOpcode& op = bco(i);
            if (op.major == IntOpcode::maJump && (op.minor & ~IntOpcode::jSymbolic) == IntOpcode::jCatch) {
                ehState = Yes;
                break;
            }
        }
    }
    return (ehState == Yes);
}

/**************************** Other Functions ****************************/

/** Check for comparison instruction. */
static bool
isComparison(const IntOpcode& op)
{
    return op.is(IntOpcode::maBinary)
        && (op.minor >= biCompareEQ && op.minor <= biCompareGT_NC);
}

/** Check for regular conditional jump. */
static bool
isConditionalJump(const IntOpcode& op)
{
    return op.isRegularJump()
        && (op.minor & IntOpcode::jPopAlways) != 0;
}

/** Check for direct storage class, i.e. storage classes that directly
    refer to a data segment that can provide/take values with defined
    ownership semantics. */
static bool
isDirectStorageClass(const IntOpcode& op)
{
    switch (op.minor) {
     case IntOpcode::sLocal:
     case IntOpcode::sStatic:
     case IntOpcode::sShared:
     case IntOpcode::sNamedShared:
     case IntOpcode::sLiteral:
        return true;

     default:
        return false;
    }
}

/** Perform instruction fusion.
    Processes the whole BCO and fuses instructions where possible.
    \param bco Byte code object, updated in-place */
void
fuseInstructions(IntBytecodeObject& bco)
{
    /* Pathological case */
    const IntBytecodeObject::pc_t n = bco.getNumInstructions();
    if (n == 0) {
        return;
    }

    IntLocalTracer tracer(bco);

    /* Iterate backwards. We'll be combining instruction pairs. */
    for (IntBytecodeObject::pc_t i = n-1; i > 0; --i) {
        const IntOpcode& me = bco(i);
        IntOpcode& prev = bco(i-1);
        switch (me.major) {
         case IntOpcode::maBinary:
            /* push + binary -> fusedbinary */
            if (prev.major == IntOpcode::maPush && isDirectStorageClass(prev)) {
                prev.major = IntOpcode::maFusedBinary;
            }
            break;

         case IntOpcode::maUnary:
            /* push + unary -> fusedunary/inplaceunary */
            if (prev.major == IntOpcode::maPush && isDirectStorageClass(prev)) {
                if (prev.minor == IntOpcode::sLocal
                    && (me.minor == unInc || me.minor == unDec)
                    && tracer.isOverwrittenLocal(i+1, prev.arg, STORE_DEPTH))
                {
                    prev.major = IntOpcode::maInplaceUnary;
                } else {
                    prev.major = IntOpcode::maFusedUnary;
                }
            }
            break;

         case IntOpcode::maJump:
            /* bcmpXX + jXXp -> fusedcomparison */
            if (isConditionalJump(me) && isComparison(prev)) {
                prev.major = IntOpcode::maFusedComparison;
            }
            break;

         case IntOpcode::maFusedComparison:
            /* push + fusedcomparison -> fusedcomparison2 */
            if (prev.major == IntOpcode::maPush && isDirectStorageClass(prev)) {
                prev.major = IntOpcode::maFusedComparison2;
            }
            break;
        }
    }
}

/** Undo instruction fusion.
    Restores the original instructions in the whole BCO.
    \param bco Byte code object, updated in-place */
void
unfuseInstructions(IntBytecodeObject& bco)
{
    for (IntBytecodeObject::pc_t i = 0, n = bco.getNumInstructions(); i < n; ++i) {
        bco(i).major = bco(i).getExternalMajor();
    }
}
