/**
  *  \file int/names.cc
  *  \brief Interpreter: Variable Name Mapping
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/names.h"
#include "io/streams.h"
#include "io/io.h"
#include "util/except.h"
#include "util/recode.h"

static const uint_t HASH_MAX = 64;

// Just a random off-the-hat hash function
static uint_t
hash(const char*const data, const size_t size)
{
    uint_t h = size;
    uint_t limit = size;
    if (limit > 5) {
        limit = 5;
    }
    for (uint_t i = 0, j = size-1; i < limit; ++i, --j) {
        h += uint_t(data[i]);
        h += uint_t(data[j]);
    }
    return h % HASH_MAX;
}

static uint_t
hash(const string_t& name)
{
    return hash(name.data(), name.size());
}

IntNameQuery::IntNameQuery(const string_t& tmp)
    : data(tmp.data()),
      size(tmp.size()),
      hash(::hash(data, size))
{ }

IntNameQuery::IntNameQuery(const char* str)
    : data(str),
      size(std::strlen(str)),
      hash(::hash(data, size))
{ }

IntNameQuery::IntNameQuery(const IntNameQuery& other, size_t pos)
    : data(other.data + pos),
      size(other.size - pos),
      hash(::hash(data, size))
{ }

/**************************** IntVariableNames ***************************/

const IntVariableNames::index_t IntVariableNames::nil;

/** Constructor. Build blank mapping. */
IntVariableNames::IntVariableNames()
    : hashFirst(HASH_MAX, nil),
      hashNext()
{ }

IntVariableNames::~IntVariableNames()
{ }

/** Look up name.
    \param name Variable name
    \return Index, nil if not found */
IntVariableNames::index_t
IntVariableNames::lookup(const IntNameQuery& q) const
{
    index_t i = hashFirst[q.getHash()];
    while (i != nil && !q.match(names[i])) {
        i = hashNext[i];
    }
    return i;
}

/** Add a mapping. Note that this does not attempt to avoid duplicates.
    \param name Variable name
    \return Assigned index */
IntVariableNames::index_t
IntVariableNames::add(const string_t& name)
{
    index_t h = hash(name);
    index_t slot = names.size();
    names.push_back(name);
    hashNext.push_back(hashFirst[h]);
    hashFirst[h] = slot;
    return slot;
}

/** Add a mapping unless it is already there.
    \param name Variable name
    \return Assigned index */
IntVariableNames::index_t
IntVariableNames::addMaybe(const string_t& name)
{
    index_t i = lookup(name);
    if (i == nil) {
        i = add(name);
    }
    return i;
}

/** Store name list to stream.
    \param s Stream to save to
    \param count Number of names to save. Must be <= getNumNames(). */
void
IntVariableNames::save(Stream& s, uint32_t count) const
{
    ASSERT(count <= names.size());
    for (uint32_t i = 0; i < count; ++i) {
        storePascalStringT(s, convertUtf8ToGame(names[i]));
    }
}

/** Load name list from stream.
    \param s Stream to load from
    \param count Number of names to load */
void
IntVariableNames::load(Stream& s, uint32_t count)
{
    while (count--) {
        string_t name;
        getPascalString(s, name);
        add(convertGameToUtf8(name));
    }
}
