/**
  *  \file ui/dialogs/filedlg.cc
  *  \brief File Chooser Dialog
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/dialogs/filedlg.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets/dircrumbs.h"
#include "ui/widgets/filechoose.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/statictext.h"
#include "util/nls.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/button.h"

/** Constructor.
    \param title Window title
    \param initial_directory Initial directory to display
    \param wildcard Initial wildcard to use */
UIFileChooserDialog::UIFileChooserDialog(string_t title, Ptr<Directory> initial_directory, string_t wildcard)
    : UIWindow(title, BLUE_WINDOW, COLOR_GRAY)
{
    init(initial_directory, wildcard);
}

UIFileChooserDialog::~UIFileChooserDialog()
{ }

void
UIFileChooserDialog::addPattern(string_t wildcard)
{
    file_chooser->addPattern(wildcard);
}

void
UIFileChooserDialog::rescan()
{
    file_chooser->rescan();
}


/** Initialize. */
void
UIFileChooserDialog::init(Ptr<Directory> initial_directory, string_t wildcard)
{
    // VBox
    //   HBox
    //     "File:"
    //     UIInputLine
    //   UIFrameGroup, VBox
    //     UIFileChooser
    //     UIScrollbar
    //   UIDirectoryCrumbTrail
    //   HBox
    //     "OK"
    //     "Cancel"
    //     UISpacer
    setLayout(UIVBoxLayout::instance5);

    file_chooser = &h.add(new UIFileChooser(1001, initial_directory, wildcard, 2, 15));
    input_line   = &h.add(new UIInputLine(500, 20, ss_Alt + 'f', UIInputLine::ilf_Framed, 1000));
    crumb_trail  = &h.add(new UIDirectoryCrumbTrail());

    UIGroup&      g1 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIFrameGroup& g2 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    UIGroup&      g4 = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(h.add(new UIStaticText(_("File:"), UIStaticText::StaticTitle)));
    g1.add(*input_line);
    add(g1);

    g2.add(*file_chooser);
    g2.add(h.add(new UIScrollbar(*file_chooser)));
    add(g2);

    add(*crumb_trail);

    g4.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addCall(this, &UIFileChooserDialog::onEnter));
    g4.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    g4.add(h.add(new UISpacer()));
    add(g4);

    add(h.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Tab)));
    pack();

    crumb_trail->setDirectory(initial_directory);
    file_chooser->sig_file_selected.add(this, &UIFileChooserDialog::onFinish);
    file_chooser->sig_directory_changed.add(this, &UIFileChooserDialog::onDirChange);
    crumb_trail->sig_changed.add(this, &UIFileChooserDialog::onCrumbClick);
}

bool
UIFileChooserDialog::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && event.key.code == SDLK_DOWN && input_line->hasState(st_Focused)) {
        file_chooser->setFocusedWidget();
        return true;
    } else if (event.type == ev_Key && event.key.code == vk_Quit) {
        stop(0);
        putKey(vk_Quit);
        return true;
    } else {
        return UIWindow::handleEvent(event, second_pass);
    }
}

/** "OK" or "Enter": process selection. */
void
UIFileChooserDialog::onEnter()
{
    if (input_line->hasState(st_Focused)) {
        /* Input line has focus. Give it to file chooser. */
        if (file_chooser->handleUserInput(input_line->getText())) {
            input_line->setText(string_t());
            file_chooser->setFocusedWidget();
            file_chooser->setSelectedWidget();
        }
    } else {
        /* File chooser has focus */
        file_chooser->handleEnter();
    }
}

/** Finish. Called by UIFileChooser when a file has been selected. */
void
UIFileChooserDialog::onFinish()
{
    stop(1);
}

/** Crumb click. Called by the UIDirectoryCrumbTrail when a crumb has
    been clicked. */
void
UIFileChooserDialog::onCrumbClick(int n)
{
    /* If last item has been clicked, do nothing, as this would mean
       no directory change. */
    if (n != crumb_trail->getNumberOfItems()-1) {
        file_chooser->setDirectory(crumb_trail->getItem(n));
        file_chooser->setFocusedWidget();
    }
}

/** Directory change. Called by UIFileChooser when the directory was
    changed by the user (or by onCrumbClick). */
void
UIFileChooserDialog::onDirChange()
{
    crumb_trail->setDirectory(file_chooser->getCurrentDirectory());
}

/** Get result file. Valid iff the dialog ran successfully with
    nonzero result. */
const string_t&
UIFileChooserDialog::getResultFile() const
{
    return file_chooser->getResultFile();
}

/** Access directory change signal. Use this to observe the dialog. */
Signal0&
UIFileChooserDialog::getDirectoryChangeSignal()
{
    return file_chooser->sig_directory_changed;
}

/** Get current directory. */
Ptr<Directory>
UIFileChooserDialog::getCurrentDirectory() const
{
    return file_chooser->getCurrentDirectory();
}
